# -*- coding: utf-8 -*-
"""
Detector shifts finding classes, using tomo-consistency.

@author: Nicola VIGANÒ, Computational Imaging group, CWI, The Netherlands,
and ESRF - The European Synchrotron, Grenoble, France
"""

import numpy as np

from typing import Callable, Sequence, Tuple, Dict, Optional, Union, List
from numpy.typing import ArrayLike, NDArray

from tqdm.auto import tqdm

from .. import solvers
from .. import projectors
from .. import models
from .. import data_terms
from .. import regularizers
from .. import attenuation
from ..processing import post as post_proc

from . import shifts
from . import centering

from dataclasses import dataclass, field

import copy as cp

import matplotlib.pyplot as plt


eps = np.finfo(np.float32).eps


NDArrayFloat = NDArray[np.floating]


def _plan_angle_groups(angles: NDArrayFloat, bunch_size: int, strategy: str) -> Sequence:
    if strategy.lower() == "random":
        angle_inds = np.random.permutation(len(angles))
    elif strategy.lower() == "equispaced":
        angle_inds = np.arange(len(angles))
    else:
        raise ValueError(f"Unknown strategy: {strategy}")

    stride = int(np.ceil(len(angles) / bunch_size))
    return [list(angle_inds[ii::stride]) for ii in range(stride)]


def get_shifts_angle(shifts_vu_ref: NDArrayFloat, shifts_vu_oth: NDArrayFloat) -> float:
    """Compute the angle between two different sets of shifts.

    Parameters
    ----------
    shifts_vu_ref : NDArrayFloat
        Reference shifts.
    shifts_vu_oth : NDArrayFloat
        Other shifts.

    Returns
    -------
    float
        The angle between the shift hypervectors.
    """
    dot_prod = shifts_vu_ref.flatten().dot(shifts_vu_oth.flatten())
    norm_ref = np.linalg.norm(shifts_vu_ref.flatten())
    norm_oth = np.linalg.norm(shifts_vu_oth.flatten())
    return np.arccos(dot_prod / (norm_oth * norm_ref))


class ReconstructionResiduals:
    """Manipulate reconstruction residuals."""

    residuals_w: NDArrayFloat

    def __init__(
        self, data_vwu: NDArrayFloat, proj_vwu: NDArrayFloat, norm: data_terms.DataFidelityBase = data_terms.DataFidelity_l2()
    ) -> None:
        """
        Compute and store the angle residuals of the projected reconstruction.

        Parameters
        ----------
        data_vwu : NDArrayFloat
            The acquired data.
        proj_vwu : NDArrayFloat
            The projected reconstruction.
        norm : DataFidelityBase, optional
            The norm to use. The default is DataFidelity_l2().
        """
        norm = cp.deepcopy(norm)
        norm.assign_data(data_vwu)
        data_residual = norm.compute_residual(proj_vwu)

        num_angles = data_residual.shape[-2]
        self.residuals_w = np.empty((num_angles,), dtype=data_residual.dtype)
        for ii in range(num_angles):
            self.residuals_w[ii] = norm[..., ii, :].compute_residual_norm(data_residual[..., ii, :])

    @property
    def norm(self) -> float:
        """
        Compute the norm of the residuals.

        Returns
        -------
        float
            The norm.
        """
        return float(np.linalg.norm(self.residuals_w))

    @property
    def mean(self) -> float:
        """
        Compute the mean of the residuals.

        Returns
        -------
        float
            The mean residual.
        """
        return self.residuals_w.mean()

    @property
    def std(self) -> float:
        """
        Compute the standard-deviation of the residuals.

        Returns
        -------
        float
            The residuals std.
        """
        return self.residuals_w.std()


@dataclass
class FittingInfo:
    """Fitting information, tracking each iteration."""

    shifts_pre: NDArrayFloat
    residual_0: ReconstructionResiduals

    shifts_ivu: List[NDArrayFloat] = field(default_factory=lambda: [])
    residuals: List[ReconstructionResiduals] = field(default_factory=lambda: [])

    fsc_ints: List[float] = field(default_factory=lambda: [])

    def __post_init__(self):
        """Initialize the FittingInfo object."""
        self.shifts_pre = self.shifts_pre.copy()

    @property
    def iterations(self) -> int:
        """Return the number of iterations."""
        return len(self.shifts_ivu)

    def get_last_residual(self) -> Tuple[ReconstructionResiduals, float]:
        """Return the last added residual (corresponding to last iteration).

        Returns
        -------
        Tuple[ReconstructionResiduals, float]
            The residual object and the normalized residual value.
        """
        return self.residuals[-1], self.residuals[-1].norm / self.residual_0.norm

    def push_iteration_result(
        self, shifts_vu: Union[ArrayLike, NDArrayFloat], residual: ReconstructionResiduals, fsc_int: Optional[float] = None
    ) -> None:
        """Add the shifts and residual from the latest iterations.

        Parameters
        ----------
        shifts_vu : ArrayLike | NDArrayFloat
            The latest shifts.
        residual : ReconstructionResiduals
            The residual of the latest iteration.
        fsc_int : float | None
            The FSC/FRC integral of the latest iteration.
        """
        self.shifts_ivu.append(np.array(shifts_vu).copy())
        self.residuals.append(residual)
        if fsc_int is not None:
            self.fsc_ints.append(fsc_int)

    def get_residuals_norm(self) -> NDArrayFloat:
        """Return all the normalized residual norms.

        Returns
        -------
        NDArrayFloat
            The normalized residual norms from all the iterations.
        """
        residuals = np.ones(self.iterations + 1)
        res_0_norm = self.residual_0.norm
        for ii in range(self.iterations):
            residuals[ii + 1] = self.residuals[ii].norm / res_0_norm
        return residuals

    def get_errors_norm(self, shifts_vu_ref: Union[ArrayLike, NDArrayFloat], norm_ord: int = 2) -> NDArrayFloat:
        """Compute the error norm of the shifts of all the iterations, with respect to reference shifts.

        This is used in synthetic test cases.

        Parameters
        ----------
        shifts_vu_ref : ArrayLike | NDArrayFloat
            The reference shifts.
        norm_ord : int, optional
            The power of the norm to use, by default 2.

        Returns
        -------
        NDArrayFloat
            The error norms.
        """
        shifts_vu_ref = np.array(shifts_vu_ref, ndmin=1)

        errors = np.zeros(self.iterations + 1)
        errors[0] = np.linalg.norm(shifts_vu_ref - self.shifts_pre, ord=norm_ord)
        for ii in range(self.iterations):
            errors[ii + 1] = np.linalg.norm(shifts_vu_ref - self.shifts_ivu[ii], ord=norm_ord)
        return errors

    def get_errors_stats(self, shifts_vu_ref: Union[ArrayLike, NDArrayFloat]) -> Tuple[NDArrayFloat, NDArrayFloat]:
        """Compute the mean and std of shifts error for all the iterations, with respect to reference shifts.

        This is used in synthetic test cases.

        Parameters
        ----------
        shifts_vu_ref : ArrayLike | NDArrayFloat
            The reference shifts.

        Returns
        -------
        Tuple[NDArrayFloat, NDArrayFloat]
            The error mean and std.
        """
        shifts_vu_ref = np.array(shifts_vu_ref, ndmin=1)

        err_mean = np.zeros(self.iterations + 1)
        err_std = np.zeros(self.iterations + 1)

        diffs = np.abs(shifts_vu_ref - self.shifts_pre)
        err_mean[0] = diffs.mean()
        err_std[0] = diffs.std()
        for ii in range(self.iterations):
            diffs = np.abs(shifts_vu_ref - self.shifts_ivu[ii])
            err_mean[ii + 1] = diffs.mean()
            err_std[ii + 1] = diffs.std()
        return err_mean, err_std

    def get_update(self, ind: int) -> NDArrayFloat:
        """Compute the shift update of a given iteration.

        Parameters
        ----------
        ind : int
            Index of the requested iteration.

        Returns
        -------
        NDArrayFloat
            The update to the shifts.
        """
        if ind == 0:
            return self.shifts_ivu[ind] - self.shifts_pre
        else:
            return self.shifts_ivu[ind] - self.shifts_ivu[ind - 1]

    def plot_fitting_progression(self) -> None:
        """Plot the fitting progression of the alignment procedure."""
        updates = [self.get_update(ii) for ii in range(self.iterations)]

        upd_lengths = np.array([np.linalg.norm(u) for u in updates])
        upd_angles_rad = np.zeros(self.iterations)

        for ii in range(1, self.iterations):
            upd_angles_rad[ii] = get_shifts_angle(updates[ii - 1], updates[ii])
        upd_angles_cum_rad = np.cumsum(upd_angles_rad)

        f, ax = plt.subplots(1, 3, figsize=[9, 3])

        point_i = np.array([0.0, 0.0])

        for ii, (a, l) in enumerate(zip(upd_angles_cum_rad, upd_lengths)):
            point_d = np.array([np.cos(a), np.sin(a)]) + l
            # print(np.rad2deg(a), l, point_i, point_d)

            ax[0].arrow(*point_i, *point_d, color=f"C{ii}")

            point_i += point_d

        ax[1].plot(upd_lengths, label="Update length")
        ax[1].grid()
        ax[1].legend()

        ax[2].semilogy(self.get_residuals_norm(), label="Residuals")
        ax[2].grid()
        ax[2].legend()

        f.tight_layout()
        plt.show(block=False)


class DetectorShiftsTC(shifts.DetectorShiftsXC):
    """Compute the detector shifts for a given dataset, through tomo-consistency."""

    def __init__(
        self,
        data_dvwu: NDArrayFloat,
        rot_angle_rad: Union[ArrayLike, NDArrayFloat],
        *,
        data_format: str = "dvwu",
        borders_dvwu: Dict = {"d": None, "v": 4, "w": None, "u": 20},
        vol_geom: Optional[models.VolumeGeometry] = None,
        solver_cls: Callable = solvers.SIRT,
        data_weights: Optional[NDArrayFloat] = None,
        beam_profile: Optional[NDArrayFloat] = None,
        attenuation_vol: Optional[attenuation.AttenuationVolume] = None,
        solver_opts: Dict = dict(lower_limit=0.0),
        solver_iterations: int = 100,
        solver_regularizer: Optional[regularizers.BaseRegularizer] = None,
        max_shifts: Union[float, NDArrayFloat, None] = None,
        compute_fsc_ints: bool = False,
        precision_decimals: int = 2,
        progress_bar: bool = True,
        verbose: bool = True,
    ):
        super().__init__(
            data_dvwu=data_dvwu,
            rot_angle_rad=rot_angle_rad,
            data_format=data_format,
            borders_dvwu=borders_dvwu,
            max_shifts=max_shifts,
            precision_decimals=precision_decimals,
            verbose=verbose,
        )

        if vol_geom is None:
            vol_geom = models.VolumeGeometry.get_default_from_data(data_dvwu, data_format=data_format)
        self.vol_geom = vol_geom

        if self.vol_geom.is_3D():
            geom_type = "3d"
        else:
            geom_type = "2d"
        self.prj_geom = models.ProjectionGeometry.get_default_parallel(geom_type=geom_type)

        self.solver_cls = solver_cls
        self.data_weights = data_weights
        self.beam_profile = beam_profile
        self.attenuation_vol = attenuation_vol
        self.solver_opts = solver_opts
        self.solver_iterations = solver_iterations
        self.solver_regularizer = solver_regularizer

        self.progress_bar = progress_bar or verbose
        self.compute_fsc_ints = compute_fsc_ints

    def _initialize_shifts(self, shifts_vu_init: Union[ArrayLike, NDArrayFloat], rand_init_std: float = 0.0) -> NDArrayFloat:
        dtype = np.float32
        shifts_vu_init = np.array(shifts_vu_init, ndmin=1, dtype=dtype)
        if len(shifts_vu_init) == 1:
            shifts_vu_fit = np.ones_like(self.angles_rad, dtype=dtype) * shifts_vu_init
        else:
            shifts_vu_fit = shifts_vu_init.copy()
        shifts_vu_fit = np.array(shifts_vu_fit, ndmin=(1 + self.vol_geom.is_3D()), dtype=dtype)

        if self.vol_geom.is_3D() and shifts_vu_fit.shape[-2] == 1:
            shifts_vu_fit = np.concatenate((np.zeros_like(shifts_vu_fit, dtype=dtype), shifts_vu_fit), axis=-2)

        # Random misplacement of shifts, to get out of local minima
        if rand_init_std > 0.0:
            rand_shifts_u = np.random.randn(len(self.angles_rad)) * rand_init_std
            if self.data_dims == 3:
                shifts_vu_fit[-1, :] += rand_shifts_u
            else:
                shifts_vu_fit += rand_shifts_u
        shifts_vu_fit = np.around(shifts_vu_fit, decimals=self.decimals)

        return shifts_vu_fit

    def _reconstruct(self, shifts_vu: Union[ArrayLike, NDArrayFloat]) -> Tuple[NDArrayFloat, NDArrayFloat]:
        shifts_vu = np.array(shifts_vu, ndmin=1)

        if self.data_weights is not None:
            data_term = data_terms.DataFidelity_wl2(self.data_weights)
        else:
            data_term = data_terms.DataFidelity_l2()
        solver = self.solver_cls(data_term=data_term, regularizer=self.solver_regularizer)

        if self.attenuation_vol is not None:
            att_args = self.attenuation_vol.get_projector_args()
            projector = projectors.ProjectorAttenuationXRF
        else:
            att_args = dict()
            projector = projectors.ProjectorUncorrected

        self.prj_geom.set_detector_shifts_vu(shifts_vu)
        with projector(self.vol_geom, self.angles_rad, prj_geom=self.prj_geom, psf=self.beam_profile, **att_args) as A:
            rec, _ = solver(A, self.data_vwu, iterations=self.solver_iterations, **self.solver_opts)
            proj_vwu = A(rec)

        return rec, proj_vwu

    def _compute_residual(self, proj_vwu: NDArrayFloat) -> ReconstructionResiduals:
        if self.data_weights is not None:
            data_term_slicing = data_terms.DataFidelity_wl2(self.data_weights[self.slicing])
        else:
            data_term_slicing = data_terms.DataFidelity_l2()

        return ReconstructionResiduals(self.data_vwu[self.slicing], proj_vwu[self.slicing], norm=data_term_slicing)

    def _update_shifts_inds(
        self, shifts_vu: Union[ArrayLike, NDArrayFloat], angles_inds: Union[ArrayLike, NDArray[np.integer]]
    ) -> NDArrayFloat:
        shifts_vu = np.array(shifts_vu, ndmin=1)
        angles_inds = np.array(angles_inds, ndmin=1, dtype=int)

        angles_othr = np.delete(self.angles_rad, angles_inds)
        angles_curr = self.angles_rad[angles_inds]

        shifts_vu_othr = np.delete(shifts_vu, angles_inds, axis=-1)
        shifts_vu_curr = shifts_vu[..., angles_inds]

        data_vwu_othr = np.delete(self.data_vwu, angles_inds, axis=-2)
        data_vwu_curr = self.data_vwu[..., angles_inds, :]

        if self.data_weights is not None:
            data_weights_othr = np.delete(self.data_weights, angles_inds, axis=-2)
            data_term = data_terms.DataFidelity_wl2(data_weights_othr)
        else:
            data_term = data_terms.DataFidelity_l2()
        solver = self.solver_cls(data_term=data_term, regularizer=self.solver_regularizer)

        if self.attenuation_vol is not None:
            other_inds = list(np.delete(np.arange(len(self.angles_rad)), angles_inds))
            att_args_othr = self.attenuation_vol.get_projector_args(rot_ind=other_inds)
            att_args_curr = self.attenuation_vol.get_projector_args(rot_ind=angles_inds)
            ProjType = projectors.ProjectorAttenuationXRF
        else:
            att_args_othr = att_args_curr = dict()
            ProjType = projectors.ProjectorUncorrected

        self.prj_geom.set_detector_shifts_vu(shifts_vu_othr)
        with ProjType(self.vol_geom, angles_othr, prj_geom=self.prj_geom, psf=self.beam_profile, **att_args_othr) as prj:
            rec, _ = solver(prj, data_vwu_othr, iterations=self.solver_iterations, **self.solver_opts)

        self.prj_geom.set_detector_shifts_vu(shifts_vu_curr)
        with ProjType(self.vol_geom, angles_curr, prj_geom=self.prj_geom, psf=self.beam_profile, **att_args_curr) as prj:
            proj_vwu_curr = prj(rec)
            if len(angles_inds) == 1 and proj_vwu_curr.ndim < data_vwu_curr.ndim:
                proj_vwu_curr = proj_vwu_curr[..., None, :]

        return self.find_shifts_vu(data_vwu_curr, proj_vwu_curr)

    def fit_vu_traditional(
        self,
        shifts_vu_init: Union[ArrayLike, NDArrayFloat],
        *,
        iterations: int = 3,
        relaxation: float = 1.0,
        regularizer: Optional[regularizers.BaseRegularizer] = None,
        ref_com_zyx: Union[ArrayLike, NDArrayFloat, None] = None,
    ) -> Tuple[NDArrayFloat, FittingInfo]:
        """Implement the traditional tomo-consistency alignment routine.

        Parameters
        ----------
        shifts_vu_init : ArrayLike | NDArrayFloat
            The initial shifts.
        iterations : int, optional
            Number of optimization iterations, by default 3
        relaxation : float, optional
            The relaxation factor for each shift update, by default 1.0
        regularizer : BaseRegularizer | None, optional
            A regularizer for the shifts, by default None
        ref_com_zyx : ArrayLike | NDArrayFloat | None, optional
            The reference center-of-mass of the reconstruction, by default None

        Returns
        -------
        Tuple[NDArrayFloat, FittingInfo]
            The fitted shifts and a FittingInfo object.
        """
        return self.fit_vu_gradient(
            shifts_vu_init=shifts_vu_init,
            iterations=iterations,
            relaxation=relaxation,
            regularizer=regularizer,
            ref_pos_zyx=ref_com_zyx,
        )

    def fit_vu_sirt(
        self,
        shifts_vu_init: Union[ArrayLike, NDArrayFloat],
        *,
        iterations: int = 3,
        relaxation: float = 1.0,
        regularizer: Optional[regularizers.BaseRegularizer] = None,
        ref_pos_zyx: Union[ArrayLike, NDArrayFloat, None] = None,
        bunch_size: int = 5,
    ) -> Tuple[NDArrayFloat, FittingInfo]:
        if bunch_size < 1:
            raise ValueError("Argument `bunch_size` needs to be >= 1.")

        return self.fit_vu_gradient(
            shifts_vu_init=shifts_vu_init,
            iterations=iterations,
            relaxation=relaxation,
            regularizer=regularizer,
            ref_pos_zyx=ref_pos_zyx,
            bunch_size=bunch_size,
        )

    def fit_vu_gradient(
        self,
        shifts_vu_init: Union[ArrayLike, NDArrayFloat],
        *,
        iterations: int = 3,
        relaxation: float = 1.0,
        regularizer: Optional[regularizers.BaseRegularizer] = None,
        ref_pos_zyx: Union[ArrayLike, NDArrayFloat, None] = None,
        bunch_size: int = 0,
        bunch_strategy: str = "equispaced",
    ) -> Tuple[NDArrayFloat, FittingInfo]:
        shifts_vu_fit = self._initialize_shifts(shifts_vu_init)

        if regularizer is not None:
            tau = 1.0 + regularizer.initialize_sigma_tau(np.ones_like(shifts_vu_fit))
            tau = relaxation / tau
        else:
            tau = relaxation

        rec, proj_vwu = self._reconstruct(shifts_vu_fit)

        recenter = centering.RecenterVolume(self.prj_geom, self.angles_rad)

        if ref_pos_zyx is None:
            ref_pos_zyx = post_proc.com(rec)
        else:
            ref_pos_zyx = np.squeeze(np.array(ref_pos_zyx))
            if ref_pos_zyx.ndim == 1:
                shifts_vu_fit = recenter.to_com(shifts_vu_fit, rec, ref_pos_zyx)
            else:
                shifts_vu_fit = recenter.as_reference(shifts_vu_fit, rec, ref_pos_zyx, method="xc")
            rec_tmp = rec
            rec, proj_vwu = self._reconstruct(shifts_vu_fit)

            if self.verbose:
                print(f"Center-of-mass: {post_proc.com(rec_tmp)} => {post_proc.com(rec)}, ref: {ref_pos_zyx}")

        res = self._compute_residual(proj_vwu)
        if self.compute_fsc_ints:
            fsc_int, _, _ = self.compute_fsc_integral(shifts_vu_fit)
        else:
            fsc_int = None
        info_fit = FittingInfo(shifts_pre=shifts_vu_fit, residual_0=res, fsc_ints=[fsc_int] if fsc_int is not None else [])

        is_SIRT_like = bunch_size > 0
        desc_top = f"Tomo-consistency ({'SIRT' if is_SIRT_like else 'Trad.'})"
        display_in_pbar = self.verbose and self.progress_bar
        for ii_o in tqdm(range(iterations), disable=((is_SIRT_like and self.verbose) or not self.progress_bar), desc=desc_top):
            # Gradient computation of the shifts
            if is_SIRT_like:
                angle_groups = _plan_angle_groups(self.angles_rad, bunch_size, strategy=bunch_strategy)

                shifts_vu_upd = np.empty_like(shifts_vu_fit)
                for iis_g in tqdm(angle_groups, disable=(not display_in_pbar), desc=f"TC-SIRT {ii_o+1}/{iterations}"):
                    shifts_vu_upd[..., iis_g] = -self._update_shifts_inds(shifts_vu_fit, iis_g)
            else:
                shifts_vu_upd = -self.find_shifts_vu(self.data_vwu, proj_vwu)

            if regularizer is not None:
                q = regularizer.initialize_dual()
                regularizer.update_dual(q, shifts_vu_fit)
                regularizer.apply_proximal(q)

                shifts_vu_upd -= regularizer.compute_update_primal(q)

            shifts_vu_fit += tau * shifts_vu_upd

            rec_tmp, proj_vwu = self._reconstruct(shifts_vu_fit)
            if ref_pos_zyx.ndim == 1:
                shifts_vu_fit = recenter.to_com(shifts_vu_fit, rec, ref_pos_zyx)
            else:
                shifts_vu_fit = recenter.as_reference(shifts_vu_fit, rec, ref_pos_zyx, method="xc")
            rec, proj_vwu = self._reconstruct(shifts_vu_fit)

            res = self._compute_residual(proj_vwu)
            if self.compute_fsc_ints:
                fsc_int, _, _ = self.compute_fsc_integral(shifts_vu_fit)
            info_fit.push_iteration_result(shifts_vu_fit, res, fsc_int)

            if self.verbose:
                print(post_proc.com(rec_tmp), "=>", post_proc.com(rec))

                res, rel_res = info_fit.get_last_residual()
                upd_vu = info_fit.get_update(ii_o)
                print(f"{res.norm} (relative: {rel_res}, update norm: {np.linalg.norm(upd_vu)})")

        if self.verbose:
            residuals = info_fit.get_residuals_norm()
            res_0_norm = info_fit.residual_0.norm
            res_last_norm = info_fit.residuals[-1].norm

            print(f"Residual: {res_0_norm}->{res_last_norm} ({residuals[-1]})")
            fig, ax = plt.subplots(figsize=[8, 2])
            ax.semilogy(residuals)
            ax.grid()
            fig.tight_layout()
            plt.show(block=False)

        return shifts_vu_fit, info_fit

    def fit_vu_sart(
        self,
        shifts_vu_init: Union[ArrayLike, NDArrayFloat],
        *,
        iterations: int = 3,
        relaxation: float = 1.0,
        ref_pos_zyx: Union[ArrayLike, NDArrayFloat, None] = None,
        bunch_size: int = 5,
        bunch_strategy: str = "random",
        rand_init_std: float = 0.0,
    ) -> Tuple[NDArrayFloat, FittingInfo]:
        shifts_vu_fit = self._initialize_shifts(shifts_vu_init, rand_init_std=rand_init_std)

        rec, proj_vwu = self._reconstruct(shifts_vu_fit)

        recenter = centering.RecenterVolume(self.prj_geom, self.angles_rad)

        if ref_pos_zyx is None:
            ref_pos_zyx = post_proc.com(rec)
        else:
            ref_pos_zyx = np.squeeze(np.array(ref_pos_zyx))
            if ref_pos_zyx.ndim == 1:
                shifts_vu_fit = recenter.to_com(shifts_vu_fit, rec, ref_pos_zyx)
            else:
                shifts_vu_fit = recenter.as_reference(shifts_vu_fit, rec, ref_pos_zyx, method="xc")
            rec_tmp = rec
            rec, proj_vwu = self._reconstruct(shifts_vu_fit)

            if self.verbose:
                print(f"Center-of-mass: {post_proc.com(rec_tmp)} => {post_proc.com(rec)}, ref: {ref_pos_zyx}")

        res = self._compute_residual(proj_vwu)
        if self.compute_fsc_ints:
            fsc_int, _, _ = self.compute_fsc_integral(shifts_vu_fit)
        else:
            fsc_int = None
        info_fit = FittingInfo(shifts_pre=shifts_vu_fit, residual_0=res, fsc_ints=[fsc_int] if fsc_int is not None else [])

        display_in_pbar = self.verbose and self.progress_bar
        for ii_o in tqdm(range(iterations), disable=(self.verbose or not self.progress_bar), desc="Tomo-consistency (SART)"):
            angle_groups = _plan_angle_groups(self.angles_rad, bunch_size, strategy=bunch_strategy)

            for iis in tqdm(angle_groups, disable=(not display_in_pbar), desc=f"TC-SART {ii_o+1}/{iterations}"):
                shifts_vu_fit[..., iis] -= relaxation * self._update_shifts_inds(shifts_vu_fit, iis)

            rec_tmp, proj_vwu = self._reconstruct(shifts_vu_fit)
            if ref_pos_zyx.ndim == 1:
                shifts_vu_fit = recenter.to_com(shifts_vu_fit, rec, ref_pos_zyx)
            else:
                shifts_vu_fit = recenter.as_reference(shifts_vu_fit, rec, ref_pos_zyx, method="xc")
            rec, proj_vwu = self._reconstruct(shifts_vu_fit)

            res = self._compute_residual(proj_vwu)
            if self.compute_fsc_ints:
                fsc_int, _, _ = self.compute_fsc_integral(shifts_vu_fit)
            info_fit.push_iteration_result(shifts_vu_fit, res, fsc_int)

            if self.verbose:
                print(post_proc.com(rec_tmp), "=>", post_proc.com(rec))

                res, rel_res = info_fit.get_last_residual()
                upd_vu = info_fit.get_update(ii_o)
                print(f"{res.norm} (relative: {rel_res}, update norm: {np.linalg.norm(upd_vu)})")

        if self.verbose:
            residuals = info_fit.get_residuals_norm()
            print(f"Residual: {info_fit.residual_0.norm}->{res.norm} ({residuals[-1]})")
            fig, ax = plt.subplots(figsize=[8, 2])
            ax.semilogy(residuals)
            ax.grid()
            fig.tight_layout()
            plt.show(block=False)

        return shifts_vu_fit, info_fit

    def compute_fsc_integral(
        self, shifts_vu: Union[ArrayLike, NDArrayFloat], full_integral: bool = False
    ) -> Tuple[float, NDArrayFloat, NDArrayFloat]:
        """Compute the integral of the FSC/FRC, with the given shifts.

        Parameters
        ----------
        shifts_vu : ArrayLike | NDArrayFloat
            The VU shift
        full_integral : bool, optional
            Whether to do the full FSC integral or just what is above the noise, by default False

        Returns
        -------
        tuple[float, NDArrayFloat, NDArrayFloat]
            The integral, the FSC and the function T
        """
        shifts_vu = np.array(shifts_vu, ndmin=1)

        angles_1 = self.angles_rad[0::2]
        angles_2 = self.angles_rad[1::2]

        shifts_vu_1 = shifts_vu[..., 0::2]
        shifts_vu_2 = shifts_vu[..., 1::2]

        prj_geom_1 = self.prj_geom[0::2]
        prj_geom_1.set_detector_shifts_vu(shifts_vu_1)
        prj_geom_2 = self.prj_geom[1::2]
        prj_geom_2.set_detector_shifts_vu(shifts_vu_2)

        if self.data_weights is not None:
            data_term_1 = data_terms.DataFidelity_wl2(self.data_weights[..., 0::2, :])
            data_term_2 = data_terms.DataFidelity_wl2(self.data_weights[..., 1::2, :])
        else:
            data_term_1 = data_term_2 = data_terms.DataFidelity_l2()
        solver_1 = self.solver_cls(data_term=data_term_1, regularizer=self.solver_regularizer)
        solver_2 = self.solver_cls(data_term=data_term_2, regularizer=self.solver_regularizer)

        if self.verbose:
            print("Computing reconstruction 1..", sep="", flush=True)

        with projectors.ProjectorUncorrected(self.vol_geom, angles_1, prj_geom=prj_geom_1, psf=self.beam_profile) as A:
            rec_1, _ = solver_1(A, self.data_vwu[..., 0::2, :], iterations=self.solver_iterations, **self.solver_opts)

        if self.verbose:
            print("\b\b: Done.\nComputing reconstruction 2..", sep="", flush=True)

        with projectors.ProjectorUncorrected(self.vol_geom, angles_2, prj_geom=prj_geom_2, psf=self.beam_profile) as A:
            rec_2, _ = solver_2(A, self.data_vwu[..., 1::2, :], iterations=self.solver_iterations, **self.solver_opts)

        if self.verbose:
            print("\b\b: Done.\nComputing FSC's integral..", sep="", flush=True)

        if self.data_dims == 3:
            slice_ind = self.data_vwu.shape[-3] // 2
            rec_1 = rec_1[slice_ind, ...]
            rec_2 = rec_2[slice_ind, ...]
        frc, Tb = post_proc.frc(rec_1, rec_2, axes=(-2, -1))

        if full_integral:
            fsc_int = np.sum(frc)
        else:
            point = post_proc.lines_intersection(frc, Tb, x_lims=(1, None))
            if point is not None:
                fsc_int = np.sum(frc[: int(point[0])])
                fsc_int += (frc[int(point[0])] + point[1]) * (point[0] - int(point[0])) / 2
            else:
                fsc_int = np.sum(frc[frc > Tb])

        if self.verbose:
            print("\b\b: Done.")

        return fsc_int, frc, Tb
